(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (factory((global.Starlight = global.Starlight || {})));
}(this, function (exports) { 'use strict';

  // ES5 output: var a=requestAnimationFrame;
  // Invocation comparison: requestAnimationFrame(b) ~~~ a(b)
  // Size delta per invocation: 24 - 4 = 20 bytes
  // Size saved after: ceil(28 / 20) = 2 usages
  var requestAnimationFrame = self.requestAnimationFrame;

  // ES5 output: var a=Array.isArray;
  // Invocation comparison: Array.isArray(b) ~~~ a(b)
  // Size delta per invocation: 16 - 4 = 12 bytes
  // Size saved after: ceil(20 / 12) = 2 usages
  var isArray = Array.isArray;

  // ES5 output: var a=Array.from;
  // Invocation comparison: Array.from(b) ~~~ a(b)
  // Size delta per invocation: 13 - 4 = 9 bytes
  // Size saved after: ceil(17 / 9) = 2 usages
  var asArray = Array.from;

  // ES5 output: var a=Object.keys;
  // Invocation comparison: Object.Keys(b) ~~~ a(b)
  // Size delta per invocation: 14 - 4 = 10 bytes
  // Size saved after: ceil(18 / 10) = 2 usages
  var ObjectKeys = Object.keys;

  // ES5 output: var a=self.document;
  // Invocation comparison: document ~~~ a
  // Size delta per invocation: 8 - 1 = 7 bytes
  // Size saved after: ceil(20 / 7) = 3 usages
  var DOMDocument = self.document;

  // ES5 output: function a(b){return"function"===typeof b};
  // Invocation comparison: typeof a==='function' ~~~ a(b)
  // Size delta per invocation: 21 - 4 = 17 bytes
  // Size saved after: ceil(43 / 17) = 3 usages
  function isFunction(obj) {
      return typeof obj === 'function';
  }

  // ES5 output: function a(b){return"string"===typeof b};
  // Invocation comparison: typeof a==='string' ~~~ a(b)
  // Size delta per invocation: 19 - 4 = 15 bytes
  // Size saved after: ceil(41 / 15) = 3 usages
  function isString(obj) {
      return typeof obj === 'string';
  }

  // ES5 output: function a(b){return"number"===typeof b};
  // Invocation comparison: typeof a==='number' ~~~ a(b)
  // Size delta per invocation: 19 - 4 = 15 bytes
  // Size saved after: ceil(41 / 15) = 3 usages
  function isNumber(obj) {
      return typeof obj === 'number';
  }

  // ES5 declaration: var a=null;
  // Invocation comparison: null ~~~ a
  // Size delta per invocation: 4 - 1 = 3 bytes
  // Size saved after: ceil(11 / 3) = 4 usages
  var Null = null;

  // ES5 output: function a(b){return void 0===b};
  // Invocation comparison: b===undefined ~~~ a(b)
  // Size delta per invocation: 13 - 4 = 9 bytes
  // Size saved after: ceil(33 / 9) = 4 usages
  function isUndefined(a) {
      return a === undefined;
  }

  // Patch Action codes
  var ADD_ACTION = 0;
  var UPDATE_ACTION = 1;
  var REMOVE_ACTION = 2;
  var REGISTER_STYLE = 4;

  // Patch Indexes
  var PATCH_TYPE = 0;
  var PATCH_ELEMENT_ID = 1;
  var PATCH_DATA = 2;
  var PATCH_RM_DATA = 3;
  var PATCH_AM_DATA = 4;

  var TEXT_NODE_VALUE = '#text';

  // Tag Indexes
  var TAG_NAME = 0;
  var TAG_ATTRS = 1;
  var TAG_CHILDREN = 2;

  // Attribute Access Types
  var ELEMENT_PROPERTY = 0;
  var ELEMENT_ATTRIBUTE = 1;
  var ELEMENT_EVENT = 2;
  var RUN_AFTER_ACTION = 3;
  var ELEMENT_LIFECYCLE_EVENT = 4;
  var LOCAL_ELEMENT_LIFECYCLE_EVENT = 5;
  var LAYOUT_REQUEST_EVENT = 6;

  var GENERIC_ERROR_MESSAGE = 'Error: see console for details';

  var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
    return typeof obj;
  } : function (obj) {
    return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
  };

  var asyncGenerator = function () {
    function AwaitValue(value) {
      this.value = value;
    }

    function AsyncGenerator(gen) {
      var front, back;

      function send(key, arg) {
        return new Promise(function (resolve, reject) {
          var request = {
            key: key,
            arg: arg,
            resolve: resolve,
            reject: reject,
            next: null
          };

          if (back) {
            back = back.next = request;
          } else {
            front = back = request;
            resume(key, arg);
          }
        });
      }

      function resume(key, arg) {
        try {
          var result = gen[key](arg);
          var value = result.value;

          if (value instanceof AwaitValue) {
            Promise.resolve(value.value).then(function (arg) {
              resume("next", arg);
            }, function (arg) {
              resume("throw", arg);
            });
          } else {
            settle(result.done ? "return" : "normal", result.value);
          }
        } catch (err) {
          settle("throw", err);
        }
      }

      function settle(type, value) {
        switch (type) {
          case "return":
            front.resolve({
              value: value,
              done: true
            });
            break;

          case "throw":
            front.reject(value);
            break;

          default:
            front.resolve({
              value: value,
              done: false
            });
            break;
        }

        front = front.next;

        if (front) {
          resume(front.key, front.arg);
        } else {
          back = null;
        }
      }

      this._invoke = send;

      if (typeof gen.return !== "function") {
        this.return = undefined;
      }
    }

    if (typeof Symbol === "function" && Symbol.asyncIterator) {
      AsyncGenerator.prototype[Symbol.asyncIterator] = function () {
        return this;
      };
    }

    AsyncGenerator.prototype.next = function (arg) {
      return this._invoke("next", arg);
    };

    AsyncGenerator.prototype.throw = function (arg) {
      return this._invoke("throw", arg);
    };

    AsyncGenerator.prototype.return = function (arg) {
      return this._invoke("return", arg);
    };

    return {
      wrap: function (fn) {
        return function () {
          return new AsyncGenerator(fn.apply(this, arguments));
        };
      },
      await: function (value) {
        return new AwaitValue(value);
      }
    };
  }();

  var defineProperty = function (obj, key, value) {
    if (key in obj) {
      Object.defineProperty(obj, key, {
        value: value,
        enumerable: true,
        configurable: true,
        writable: true
      });
    } else {
      obj[key] = value;
    }

    return obj;
  };

  var _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  var slicedToArray = function () {
    function sliceIterator(arr, i) {
      var _arr = [];
      var _n = true;
      var _d = false;
      var _e = undefined;

      try {
        for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
          _arr.push(_s.value);

          if (i && _arr.length === i) break;
        }
      } catch (err) {
        _d = true;
        _e = err;
      } finally {
        try {
          if (!_n && _i["return"]) _i["return"]();
        } finally {
          if (_d) throw _e;
        }
      }

      return _arr;
    }

    return function (arr, i) {
      if (Array.isArray(arr)) {
        return arr;
      } else if (Symbol.iterator in Object(arr)) {
        return sliceIterator(arr, i);
      } else {
        throw new TypeError("Invalid attempt to destructure non-iterable instance");
      }
    };
  }();

  function nop() {
      return;
  };

  function traverseThunk(thunk, errorDefault) {
      if (!isFunction(thunk)) {
          return thunk;
      }
      var result = thunk;
      do {
          try {
              result = result();
          } catch (e) {
              if (isUndefined(errorDefault)) {
                  throw e;
              }
              console.error('Error rendering ' + result.name + ':', e);
              result = errorDefault;
          }
      } while (isFunction(result));
      return result;
  }

  // From https://github.com/gaearon/react-pure-render/ (MIT)
  function shallowEqual(objA, objB) {
      if (objA === objB) {
          return true;
      }

      if ((typeof objA === 'undefined' ? 'undefined' : _typeof(objA)) !== 'object' || Null === objA || (typeof objB === 'undefined' ? 'undefined' : _typeof(objB)) !== 'object' || Null === objB) {
          return false;
      }

      var keysA = ObjectKeys(objA);
      var keysB = ObjectKeys(objB);

      if (keysA.length !== keysB.length) {
          return false;
      }

      // Test for A's keys different from B.
      var bHasOwnProperty = Object.prototype.hasOwnProperty.bind(objB);
      for (var i = 0; i < keysA.length; i++) {
          if (!bHasOwnProperty(keysA[i]) || objA[keysA[i]] !== objB[keysA[i]]) {
              return false;
          }
      }

      return true;
  }

  /*
  * memoize.js
  * by @philogb and @addyosmani
  * with further optimizations by @mathias
  * and @DmitryBaranovsk
  * perf tests: http://bit.ly/q3zpG3
  * Released under an MIT license.
  */

  function memoize(fn) {
      fn.memoize = {};

      var wrappedFn = function wrappedFn() {
          var args = asArray(arguments),
              hash = '',
              i = args.length;
          var currentArg = Null;
          while (i--) {
              currentArg = args[i];
              hash += currentArg === Object(currentArg) ? JSON.stringify(currentArg) : currentArg;
          }
          return hash in fn.memoize ? fn.memoize[hash] : fn.memoize[hash] = fn.apply(this, args);
      };

      wrappedFn.clear = function () {
          fn.memoize = {};
      };

      return wrappedFn;
  }

  var getEventFlags = memoize(function getEventFlags(eventName, name, type) {
      var flags = [];

      if (name === 'input') {
          if (type === 'checkbox') {
              flags.push('target.checked');
          } else {
              flags.push('target.value');
          }
          if (['onkeyup', 'onkeydown', 'onkeypress'].indexOf(eventName) !== -1) {
              flags.push('keyCode', 'ctrlKey', 'altKey', 'shiftKey');
          }
      } else {
          flags.push('target.value');
      }
      return flags.join(',');
  });

  function isTag(obj) {
      return isArray(obj) && isString(obj[TAG_NAME]);
  }

  /*
      Usage:

      Router(rootURL, routeSpec, onUnmatched):

      rootURL (string)
          the url prefix of all paths.

      routeSpec (object)

      onUnmatched (optional function(api, badPath, lastGoodRoute))
          Called whenever an unmatched route is navigated. Parameters are:
              api (Router instance)
              badpath (string)
              lastGoodRoute (object)
                  a route descriptor object for the last good route.

      returns an api object:
          goto(path) - match the path to a route and navigate there
          href(pathName, vars) - build a path based on the name and supplied vars
          listen() - listen to window.onhashchange for route changes
          current() - get the current route descriptor object

      Example:

      const router = Router(
          '#',
          {
              'home': {
                  path: '/',
                  onenter: (r, route) => {
                      console.log('At application home');
                  },
                  onexit: (r, route, newRoute) => {
                      console.log(`Exiting from ${route.path} to ${newRoute.path}`);
                  }
              },
              'article': {
                  path: '/article/:id',
                  vars: {id: /[a-f0-9]{6,40}/},
                  onenter: (r, route) => {
                      console.log('Opening Article', route.vars.id);
                  },
                  onexit: (r, route, newRoute) => {
                      console.log('Closing Article', route.vars.id);
                  }
              },
          },
          (r, path, lastGoodRoute) => {
              alert(`Unmatched route "${path}"`);
              r.goto(lastGoodRoute.path || 'home');
          }
      );
  */

  function Router(baseUrl, routes, unmatched) {
      var listening = false;
      var currentRoute = Null;
      var reEnterHook = Null;

      var VARMATCH_RE = /:([^\/]+)/g;
      var ROUTEELEMENT_RE = /^[^\/]+$/;

      var routeMatcher = ObjectKeys(routes).map(function (name) {
          var route = routes[name];
          var reg = route.path.replace(VARMATCH_RE, function (m, varName) {
              var varDef = route.vars || {};
              var regExStr = '' + (varDef[varName] || /[^\/]+/);
              return '(' + regExStr.substring(1, regExStr.lastIndexOf('/')) + ')';
          });

          return _extends({
              matcher: new RegExp('^' + baseUrl + reg + '$'),
              name: name
          }, route);
      });

      function goto(urlOrName, vars) {
          var url = urlOrName.startsWith(baseUrl) ? urlOrName : href(urlOrName, vars);

          if (listening && _location() !== url) {
              // If we're not there make the change and exit.
              location.hash = url;
              return;
          }

          if (currentRoute && currentRoute.path === url) {
              // This is only supposed to happen when recovering from a bad path.
              return;
          }

          var routeVars = {};

          var routeMatch = routeMatcher.find(function (_ref) {
              var matcher = _ref.matcher,
                  path = _ref.path;

              var match = url.match(matcher);
              if (!match) {
                  return false;
              }

              if (path.indexOf(':') !== -1) {
                  match.shift();
                  // $FlowIgnore
                  path.replace(VARMATCH_RE, function (_, varName) {
                      // We're abusing RegExp.replace here.
                      routeVars[varName] = match.shift();
                  });
              }

              return true;
          });

          if (routeMatch) {
              (function () {
                  var newRoute = {
                      name: routeMatch.name,
                      vars: routeVars,
                      path: url
                  };
                  if (currentRoute && currentRoute.name === newRoute.name && isFunction(reEnterHook)) {
                      // $FlowIgnore
                      reEnterHook(newRoute);
                      currentRoute = newRoute;
                  } else {
                      var onexit = currentRoute && currentRoute.name ? routes[currentRoute.name].onexit || nop : nop;
                      Promise.resolve(onexit(api, currentRoute, newRoute)).then(function () {
                          reEnterHook = routes[routeMatch.name].onenter(api, newRoute);
                          currentRoute = newRoute;
                      });
                  }
              })();
          } else if (unmatched) {
              unmatched(api, url, currentRoute);
          } else {
              if (currentRoute && listening) {
                  location.hash = currentRoute.path;
                  return;
              }
              throw new Error('No route for "' + url + '"');
          }
      };

      function href(routeName) {
          var vars = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

          var route = routes[routeName];
          if (!route) {
              throw new Error('Invalid route ' + routeName + '.');
          }

          var path = '' + route.path;

          if (route.vars) {
              path = path.replace(VARMATCH_RE, function (_, varName) {
                  var value = vars[varName];
                  if ((route.vars[varName] || ROUTEELEMENT_RE).test(value)) {
                      value = isUndefined(value) ? '' : '' + value;
                      return value;
                  }

                  throw new Error('Invalid value for route ' + path + ' var ' + varName + ': ' + value + '.');
              });
          }
          return '' + baseUrl + path;
      };

      function _location() {
          return location.hash;
      };

      function listen(initialRoute) {
          if (listening) {
              return;
          }

          self.addEventListener('hashchange', function () {
              return goto(_location());
          }, false);
          listening = true;
          goto(_location() || initialRoute);
      };

      function current() {
          return currentRoute;
      }

      var api = {
          goto: goto,
          href: href,
          listen: listen,
          current: current
      };

      return api;
  }

  var pool = [];
  var nodeCounter = 0;

  function allocateNode() {
      var node = pool.length > 0 ? pool.pop() : {
          id: Null,
          name: '',
          attrs: {},
          children: [],
          static: false,
          svg: false,
          component: Null,
          key: '',
          style: [],
          delayFree: false
      };
      node.id = ++nodeCounter;
      return node;
  }

  function freeNode(node) {
      if (node.delayFree) {
          return;
      }
      if (!node.component) {
          node.children.map(freeNode);
      }
      node.component = Null;
      node.static = node.svg = false;
      node.attrs = {};
      node.name = node.key = '';
      node.children = node.style = [];

      pool.push(node);
  }

  function isObject(obj) {
      return (typeof obj === 'undefined' ? 'undefined' : _typeof(obj)) === 'object' && Null !== obj && !isArray(obj);
  }

  function isComponent(obj) {
      return isArray(obj) && isFunction(obj[TAG_NAME]);
  }

  function shouldReplaceNode(currentNode, nextNode) {
      return nextNode.name !== currentNode.name || nextNode.component !== currentNode.component;
  }

  function processTagClass(classAttr) {
      var classObj = traverseThunk(classAttr);
      if (isObject(classObj)) {
          classObj = ObjectKeys(classObj).filter(function (key) {
              return traverseThunk(classObj[key]);
          });
      } else if (isString(classObj)) {
          classObj = classObj.split(' ');
      }
      return classObj;
  }

  var attributesParser = /(?:(^|#|\.)([^#\.\[\]]+))|(\[(.+?)(?:\s*=\s*("|'|)((?:\\["'\]]|.)*?)\5)?\])/g;

  function parseTagAttrs(signature, attrs) {
      var _parseTagSignature = parseTagSignature(signature),
          _parseTagSignature2 = slicedToArray(_parseTagSignature, 3),
          tag = _parseTagSignature2[0],
          sigAttrs = _parseTagSignature2[1],
          classes = _parseTagSignature2[2];

      var newAttrs = _extends({}, sigAttrs, {
          class: classes.concat(processTagClass(attrs['class'] || '')).concat(processTagClass(attrs.className || '')).join(' ').trim()
      });

      ObjectKeys(attrs).forEach(function (key) {
          if (key.indexOf('class') === 0) {
              return;
          }
          var value = attrs[key];
          newAttrs[key] = key.indexOf('on') === 0 || key[0] === '_' ? value : traverseThunk(value);
      });

      return [tag, newAttrs];
  };

  var parseTagSignature = memoize(function (signature) {
      var tag = 'div';
      var classes = [];
      var match = void 0;
      var attrs = {};

      while (match = attributesParser.exec(signature)) {
          if (match[1] === '' && match[2]) {
              tag = match[2];
          } else if (match[1] === '#') {
              attrs.id = match[2];
          } else if (match[1] === '.') {
              classes.push(match[2]);
          } else if (match[3][0] === '[') {
              var attrValue = match[6];
              if (attrValue) attrValue = attrValue.replace(/\\(["'])/g, '$1');
              attrs[match[4]] = attrValue || true;
          }
      }

      return [tag, attrs, classes];
  });

  function getNodeKey(node, parent, index) {
      return node.attrs._key ? (node.component.name || 'tag') + '-' + node.attrs._key : parent ? 'sl_' + parent.id.toString(36) + '_' + index.toString(36) : 'sl_root';
  }

  function convertTagToNodes(parent, tag, index) {
      // tag can be a thunk, string, array of tags or array of arrays.
      // we need to deal with each.

      // Thunk

      // FIXME - There could be an issue here where a thunk and a component could
      // be confused if the thunk is the first thing out of a render function.
      // Lock this down!
      if (isFunction(tag)) {
          var state = traverseThunk(tag, GENERIC_ERROR_MESSAGE);
          return convertTagsToNodes(parent, state);
      }

      var node = allocateNode();

      // String (Text tag)
      if (isString(tag) || isNumber(tag)) {
          node.name = TEXT_NODE_VALUE;
          node.attrs = {
              textContent: tag.toString()
          };
          node.static = true;
      } else if (isTag(tag)) {
          // Element tag

          var hasAttrs = isObject(tag[TAG_ATTRS]);
          var attrs = hasAttrs ? tag[TAG_ATTRS] : {};

          var _parseTagAttrs = parseTagAttrs(tag[TAG_NAME], attrs);

          var _parseTagAttrs2 = slicedToArray(_parseTagAttrs, 2);

          node.name = _parseTagAttrs2[0];
          node.attrs = _parseTagAttrs2[1];

          node.key = node.attrs._key || node.attrs.id || Null;

          if (parent.svg || node.name === 'svg') {
              node.svg = true;
              node.attrs._ns = 'svg';
          }
          node.style = parent.style;
          node.style.forEach(function (s) {
              node.attrs[s] = '';
          }); // Add style attributes to the tag.
          node.children = convertTagChildrenToNodes(node, tag.slice(hasAttrs ? TAG_CHILDREN : TAG_ATTRS));
      } else if (isComponent(tag)) {
          var _hasAttrs = isObject(tag[TAG_ATTRS]);
          var _attrs = _hasAttrs ? tag[TAG_ATTRS] : {};

          node.component = tag[TAG_NAME];
          node.attrs = _attrs;
          node.key = getNodeKey(node, parent, index);
          node.style = parent.style.slice();
          node.static = true;
          node.children = tag.slice(_hasAttrs ? TAG_CHILDREN : TAG_ATTRS);
      } else {
          freeNode(node);
          // Array
          return convertTagsToNodes(parent, tag);
      }

      return node;
  }

  function flatten(arr) {
      return arr.reduce(function (flat, toFlatten) {
          return flat.concat(isArray(toFlatten) && !isTag(toFlatten) ? flatten(toFlatten) : toFlatten);
      }, []);
  }

  function convertTagChildrenToNodes(parent, tags) {
      if (isUndefined(tags) || Null === tags) {
          return [];
      }
      return flatten(tags.map(function (t, i) {
          return convertTagToNodes(parent, t, i);
      }));
  }

  function convertTagsToNodes(parent, tags) {
      if (isTag(tags) || isString(tags) || isComponent(tags)) {
          return [convertTagToNodes(parent, tags, 0)];
      }
      return convertTagChildrenToNodes(parent, tags);
  }

  var eventCounter = 0;
  var existingEventMap = new Map();
  var events = new Map();
  var selectorMatcher = /(?:(?:}|(?:(?:@[^{]+){))\s*)([^,@{]+)(?:,\s*([^,@{]+))*/gm;

  function Component(node, onpatch, parent) {
      var parentContext = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};

      var state = [];
      var rawState = [];
      var pendingPatches = [];
      var checkWillUpdate = Null;
      var updateLock = false;
      var updateRequested = false;
      var instance = void 0;
      var previousProps = void 0;
      var previousChildren = [];
      var parentOffset = 0;
      var version = 0;
      // Clone the nodes so they can be freed in the parent's diff process.
      var hostNode = _extends({}, node);
      var parentNode = _extends({}, parent);
      var context = _extends({}, parentContext);
      var componentMap = new Map();

      function dispatchEvent(data) {
          var callback = events.get(data.id);
          if (callback) {
              // Flow doesn't like my PseudoEventType going into a real EventHandler. Can't imagine why.
              // $FlowIgnore
              return callback(data);
          }
      }

      function replaceEventHandlers(_node) {
          var name = _node.name,
              attrs = _node.attrs,
              id = _node.id;

          for (var key in attrs) {
              var value = attrs[key];
              if (key.startsWith('on') || key[0] === '_' && isFunction(value)) {
                  if (!existingEventMap.has(id)) {
                      existingEventMap.set(id, {});
                  }
                  var elementRecord = existingEventMap.get(id);
                  var eventId = elementRecord[key];

                  if (isUndefined(eventId)) {
                      eventId = eventCounter++;
                      elementRecord[key] = eventId;
                      events.set(eventId, value);
                  } else if (value !== events.get(eventId)) {
                      events.set(eventId, value);
                  }
                  var flagKey = '_' + key + '_flags';
                  var flags = attrs[flagKey];
                  attrs[key] = eventId;
                  if (flags) {
                      attrs[flagKey] = flags.join(',');
                  }
              }
          }

          return attrs;
      }

      function removeEventHandlers(elementId) {
          var elementRecord = existingEventMap.get(elementId);
          if (isUndefined(elementRecord)) {
              return;
          }

          for (var eventName in elementRecord) {
              events.delete(elementRecord[eventName]);
          }
          existingEventMap.delete(elementId);
      }

      function addUpdateChildComponent(parent, _node, indexOffset) {
          var key = _node.key;

          var childInstance = componentMap.get(key);
          if (!childInstance) {
              childInstance = Component(_node, onpatch, parent, context);
              componentMap.set(key, childInstance);
          }

          var _childInstance$getPat = childInstance.getPatches(_node.attrs, _node.children, indexOffset),
              _childInstance$getPat2 = slicedToArray(_childInstance$getPat, 3),
              patches = _childInstance$getPat2[0],
              childRequestsUpdate = _childInstance$getPat2[1],
              childCount = _childInstance$getPat2[2];

          updateRequested = updateRequested || childRequestsUpdate;
          return [patches, childCount];
      }

      function addNode(_node, parent, index) {
          if (_node.component) {
              return addUpdateChildComponent(parent, _node, index);
          }
          var name = _node.name,
              component = _node.component;

          var patchAttrs = _node.attrs ? replaceEventHandlers(_node) : {};
          if (patchAttrs._willInsert) {
              updateLock = true;
              dispatchEvent({
                  type: '_willInsert',
                  id: patchAttrs._willInsert
              });
              updateLock = false;
          }
          return [[[ADD_ACTION, _node.id, [component ? component.name : name, patchAttrs, parent.id, index]]].concat(_node.static ? [] : diff([], _node.children, _node, 0)), 1];
      }

      function removeNode(_node) {
          var patches = void 0;
          var processRemove = function processRemove() {
              removeEventHandlers(_node.id);
              return diff(_node.children, [], _node).concat([[REMOVE_ACTION, _node.id]]);
          };

          if (_node.component) {
              var componentInstance = componentMap.get(_node.key);
              if (componentInstance) {
                  patches = componentInstance.remove();
                  componentMap.delete(_node.key);
              }
          } else if (_node.attrs._willRemove) {
              updateLock = true;
              var thennable = dispatchEvent({
                  type: '_willRemove',
                  id: _node.attrs._willRemove
              });
              updateLock = false;
              if (thennable && isFunction(thennable.then)) {
                  _node.delayFree = true;
                  thennable.then(function () {
                      onpatch(processRemove());
                      _node.delayFree = false;
                      freeNode(_node);
                  });
                  patches = [];
              } else {
                  patches = processRemove();
              }
          } else {
              patches = processRemove();
          }
          return patches;
      }

      function diff(currentState, nextState, parent, indexOffset) {
          var patches = [];
          var currentCount = currentState.length;
          var nextCount = nextState.length;

          var index = 0;

          while (index < currentCount) {
              var nextNode = index < nextCount ? nextState[index] : Null;
              var current = currentState[index];

              if (!nextNode) {
                  patches.push.apply(patches, removeNode(current));
                  if (current.delayFree) {
                      nextState[index] = current;
                      nextNode = current;
                  }
              } else if (shouldReplaceNode(current, nextNode)) {
                  patches.push.apply(patches, removeNode(current));
                  if (current.delayFree) {
                      nextState[index] = current;
                      freeNode(nextNode);
                      nextNode = current;
                  } else {
                      var _addNode = addNode(nextNode, parent, index + indexOffset),
                          _addNode2 = slicedToArray(_addNode, 2),
                          nodePatches = _addNode2[0],
                          childCount = _addNode2[1];

                      patches.push.apply(patches, nodePatches);
                      index += childCount - 1;
                  }
              } else if (nextNode.component) {
                  // nodes are roughly equal but they're components so we compare them differently
                  nextNode.key = current.key;

                  var _addUpdateChildCompon = addUpdateChildComponent(parent, nextNode, index + indexOffset),
                      _addUpdateChildCompon2 = slicedToArray(_addUpdateChildCompon, 2),
                      _nodePatches = _addUpdateChildCompon2[0],
                      _childCount = _addUpdateChildCompon2[1];

                  patches.push.apply(patches, _nodePatches);
                  index += _childCount - 1;
              } else {
                  nextNode.id = current.id;

                  var currentAttrs = current.attrs;
                  var normalizedNextAttributes = nextNode.attrs ? replaceEventHandlers(nextNode) : {};
                  var newAttrs = {};
                  var rmAttrs = [];
                  var attrsAreEqual = true;

                  for (var key in currentAttrs) {
                      var val = normalizedNextAttributes[key];
                      if (!isUndefined(val)) {
                          if (val !== currentAttrs[key]) {
                              newAttrs[key] = val;
                              attrsAreEqual = false;
                          }
                      } else {
                          rmAttrs.push(key);
                          attrsAreEqual = false;
                      }
                  }

                  for (var _key in normalizedNextAttributes) {
                      if (isUndefined(currentAttrs[_key])) {
                          newAttrs[_key] = normalizedNextAttributes[_key];
                          attrsAreEqual = false;
                      }
                  }

                  if (!attrsAreEqual) {
                      // Update the current element
                      patches.push([UPDATE_ACTION, current.id, newAttrs, rmAttrs]);
                  }

                  if (!nextNode.static) {
                      patches.push.apply(patches, diff(current.children, nextNode.children, nextNode, 0));
                  }
              }

              index++;
          }

          while (index < nextCount) {
              var _nextNode = nextState[index];

              var _addNode3 = addNode(_nextNode, parent, index + indexOffset),
                  _addNode4 = slicedToArray(_addNode3, 2),
                  _nodePatches2 = _addNode4[0],
                  _childCount2 = _addNode4[1];

              patches.push.apply(patches, _nodePatches2);
              index += _childCount2; // - 1 + 1  -1 to correct for the count, +1 for adding this node.
          }

          return patches;
      }

      function updateChildContext(c) {
          ObjectKeys(c).forEach(function (k) {
              return context[k] = c[k];
          });
      }

      function registerStyle(css) {
          var key = getKey();
          hostNode.style.push(key);
          var outCSS = ('} ' + css.trim()).replace(selectorMatcher, function (match) {
              for (var _len = arguments.length, p = Array(_len > 1 ? _len - 1 : 0), _key2 = 1; _key2 < _len; _key2++) {
                  p[_key2 - 1] = arguments[_key2];
              }

              p.slice(0, p.length - 2).forEach(function (s) {
                  if (s) {
                      match = match.split(s).join(s.trim() + '[' + key + ']');
                  }
              });
              return match;
          }).slice(2);

          pendingPatches.push([REGISTER_STYLE, getKey(), outCSS]);
      }

      function getKey() {
          return hostNode.key;
      }

      function handleUpdate(checkFunction) {
          checkWillUpdate = checkFunction;
      }

      function update(props) {
          if (updateLock) {
              updateRequested = true;
              return;
          }

          var _getPatches2 = getPatches(props),
              _getPatches3 = slicedToArray(_getPatches2, 2),
              patches = _getPatches3[0],
              childRequestsUpdate = _getPatches3[1];

          patches.length && onpatch(patches);
          if (childRequestsUpdate) {
              redraw();
          }
      }

      function _getPatches() {
          var newState = convertTagsToNodes(hostNode, rawState);
          var patches = diff(state, newState, parentNode, parentOffset);
          if (pendingPatches.length) {
              patches.push.apply(patches, pendingPatches);
              pendingPatches = [];
          }
          state.map(freeNode);
          state = newState;
          return patches;
      }

      function versionedRedrawFactory(frozenVersion) {
          return function versionedLockout() {
              if (frozenVersion === version) {
                  redraw();
              }
          };
      }

      function getPatches(props, children, indexOffset) {
          if (isUndefined(props)) {
              props = previousProps;
          }

          if (isUndefined(children)) {
              children = previousChildren;
          } else {
              previousChildren = children;
          }

          if (!isUndefined(indexOffset)) {
              parentOffset = indexOffset;
          }

          updateLock = true;
          if (instance && checkWillUpdate && !checkWillUpdate(props, previousProps)) {
              previousProps = props;
          } else if (!instance || !shallowEqual(previousProps, props)) {
              version += 1;
              instance = hostNode.component(_extends({
                  redraw: versionedRedrawFactory(version),
                  getKey: getKey,
                  registerStyle: registerStyle,
                  handleUpdate: handleUpdate,
                  updateChildContext: updateChildContext
              }, context), props);
              previousProps = props;
          }
          updateLock = false;

          rawState = instance(children);
          var patches = _getPatches();
          var isRequestingUpdate = updateRequested;
          updateRequested = false;
          var childCount = isTag(rawState) ? 1 : rawState.length;
          return [patches, isRequestingUpdate, childCount];
      }

      function redraw() {
          if (!instance) {
              return;
          }
          update();
      }

      function remove() {
          version += 1;
          rawState = [];
          return _getPatches();
      }

      return {
          remove: remove,
          redraw: redraw,
          update: update,
          getPatches: getPatches,
          dispatchEvent: dispatchEvent
      };
  }

  var RECTANGLE_PROPS = 'height width top right bottom left x y'.split(' ');

  function injectElement(element, parent, index) {
      var nextSibling = parent.childNodes[index];
      if (nextSibling) {
          parent.insertBefore(element, nextSibling);
      } else {
          parent.appendChild(element);
      }
  }

  function callPropertyFactory(element, propName) {
      return function () {
          return element[propName]();
      };
  }

  function annotatePatchAttributes(patch) {
      if (patch[PATCH_TYPE] !== ADD_ACTION && patch[PATCH_TYPE] !== UPDATE_ACTION) {
          return patch;
      }

      var accessMap = {};
      var patchData = patch[PATCH_DATA];
      var patchAttrs = patch[PATCH_TYPE] === ADD_ACTION ? patchData[1] : patchData;
      for (var name in patchAttrs) {
          accessMap[name] = annotateAttribute(name);
      }

      if (patch[PATCH_TYPE] === UPDATE_ACTION) {
          patch[PATCH_RM_DATA].forEach(function (name) {
              accessMap[name] = annotateAttribute(name);
          });
      }

      patch.push(accessMap);

      return patch;
  }

  var ACTION_ATTRIBUTES = ['focus', 'select', 'blur', 'click'];
  var ELEMENT_PROPERTIES = 'textContent value checked disabled readOnly scrollTop scrollLeft'.split(' ');
  var LOCAL_LIFECYCLE_PROPERTIES = ['_didInsert', '_willInsert', '_willRemove'];
  var LIFECYCLE_PROPERTIES = '_didRemove';
  var LAYOUT_REQUEST = '_onLayout';
  function annotateAttribute(name) {
      if (ELEMENT_PROPERTIES.indexOf(name) !== -1) {
          return ELEMENT_PROPERTY;
      } else if (LOCAL_LIFECYCLE_PROPERTIES.indexOf(name) !== -1) {
          return LOCAL_ELEMENT_LIFECYCLE_EVENT;
      } else if (LIFECYCLE_PROPERTIES === name) {
          return ELEMENT_LIFECYCLE_EVENT;
      } else if (LAYOUT_REQUEST === name) {
          return LAYOUT_REQUEST_EVENT;
      } else if (ACTION_ATTRIBUTES.indexOf(name) !== -1) {
          return RUN_AFTER_ACTION;
      } else if (name.indexOf('on') === 0) {
          return ELEMENT_EVENT;
      }
      return ELEMENT_ATTRIBUTE;
  }

  function Renderer(domRoot) {
      var _ACTION_METHODS;

      var rootId = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
      var dispatchEvent = arguments[2];

      var elementMap = new Map();
      var eventMap = new Map();
      var styleMap = new Map();
      var pendingPatches = [];
      var paused = false;
      var postProcessing = [];
      var layoutRequests = new Map();
      var layoutValues = new Map();

      window.addEventListener('resize', maintenance, false);

      elementMap.set(rootId, domRoot);

      function lifecycleEventFactory(type, id) {
          requestAnimationFrame(function () {
              dispatchEvent({
                  type: type,
                  id: parseInt(id, 10)
              });
          });
      }

      function eventHandler(evt) {
          if (!evt instanceof Event) {
              // Lifecycle events are not DOM events.
              dispatchEvent(evt);
          }
          var eventName = evt.type;
          var dataset = evt.currentTarget.dataset;
          var id = parseInt(dataset['on' + eventName + 'EventId'], 10);
          var flags = eventMap.get(id).split(',');
          var fakeEvt = {
              id: id,
              type: eventName
          };
          var stop = true;
          var index = flags.length;

          var _loop = function _loop() {
              var flag = flags[index];
              if (flag === 'bubble') {
                  stop = false;
                  return 'continue';
              }
              var flagPaths = flag.split('.');
              var pathLastIndex = flagPaths.length - 1;
              var pointer = evt;
              var fakeEvtPointer = fakeEvt;
              flagPaths.forEach(function (attr, i) {
                  var newPointer = pointer[attr];
                  if (i !== pathLastIndex) {
                      pointer = newPointer;
                      fakeEvtPointer = fakeEvtPointer[attr] || (fakeEvtPointer[attr] = {});
                  } else if (isFunction(newPointer)) {
                      pointer[attr](); // run with context
                  } else {
                      fakeEvtPointer[attr] = newPointer;
                  }
              });
          };

          while (index--) {
              var _ret = _loop();

              if (_ret === 'continue') continue;
          }
          if (stop) {
              evt.stopPropagation();
          }
          dispatchEvent(fakeEvt);
      }

      function setAttributes(element, elementId, attrs, rmAttrs, accessMap) {
          var _iteratorNormalCompletion = true;
          var _didIteratorError = false;
          var _iteratorError = undefined;

          try {
              for (var _iterator = ObjectKeys(attrs)[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                  var name = _step.value;

                  var _value = attrs[name];

                  if (_value === Null) {
                      continue;
                  }

                  switch (accessMap[name]) {
                      case ELEMENT_PROPERTY:
                          {
                              element[name] = _value;
                              continue;
                          }
                      case ELEMENT_EVENT:
                          {
                              var _eventName = name.slice(2);
                              var eventIdAttrName = name + 'EventId';
                              if (element.dataset[eventIdAttrName]) {
                                  element.removeEventListener(_eventName, eventHandler);
                              }
                              if (_value) {
                                  element.dataset[eventIdAttrName] = _value;
                                  eventMap.set(_value, attrs['_' + name + '_flags'] || getEventFlags(_eventName, element.tagName.toLowerCase(), attrs.type));
                                  element.addEventListener(_eventName, eventHandler, false);
                              }
                              continue;
                          }
                      case ELEMENT_LIFECYCLE_EVENT:
                          {
                              element.dataset[name] = _value;
                              continue;
                          }
                      case ELEMENT_ATTRIBUTE:
                          {
                              element.setAttribute(name, _value);
                              continue;
                          }
                      case LAYOUT_REQUEST_EVENT:
                          {
                              layoutRequests.set(elementId, _value);
                              continue;
                          }
                      case RUN_AFTER_ACTION:
                          {
                              if (_value) {
                                  postProcessing.push(callPropertyFactory(element, name));
                              }
                          }
                      default:
                  }
              }
          } catch (err) {
              _didIteratorError = true;
              _iteratorError = err;
          } finally {
              try {
                  if (!_iteratorNormalCompletion && _iterator.return) {
                      _iterator.return();
                  }
              } finally {
                  if (_didIteratorError) {
                      throw _iteratorError;
                  }
              }
          }

          var _iteratorNormalCompletion2 = true;
          var _didIteratorError2 = false;
          var _iteratorError2 = undefined;

          try {
              for (var _iterator2 = rmAttrs[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                  var data = _step2.value;

                  var _data = slicedToArray(data, 2),
                      accessCode = _data[0],
                      _name = _data[1];

                  switch (accessCode) {
                      case ELEMENT_EVENT:
                          element.removeEventListener(eventName, eventHandler);
                          continue;
                      case ELEMENT_PROPERTY:
                          {
                              element[_name] = value;
                              continue;
                          }
                      case ELEMENT_ATTRIBUTE:
                          {
                              element.removeAttribute(_name, value);
                          }
                      case LAYOUT_REQUEST_EVENT:
                          {
                              layoutRequests.delete(elementId);
                              layoutValues.delete(elementId);
                              continue;
                          }

                      case ELEMENT_LIFECYCLE_EVENT:
                          {
                              element.dataset.removeProp(_name);
                              continue;
                          }
                      default:
                  }
              }
          } catch (err) {
              _didIteratorError2 = true;
              _iteratorError2 = err;
          } finally {
              try {
                  if (!_iteratorNormalCompletion2 && _iterator2.return) {
                      _iterator2.return();
                  }
              } finally {
                  if (_didIteratorError2) {
                      throw _iteratorError2;
                  }
              }
          }
      }

      function addElement(patch) {
          var elementId = patch[PATCH_ELEMENT_ID];
          var data = patch[PATCH_DATA];
          var accessMap = patch[PATCH_RM_DATA];

          var _data2 = slicedToArray(data, 4),
              name = _data2[0],
              props = _data2[1],
              parentId = _data2[2],
              index = _data2[3];

          var element = void 0;

          if (name === TEXT_NODE_VALUE) {
              // $FlowIgnore
              element = DOMDocument.createTextNode(props.textContent);
          } else {
              if (props._ns === 'svg') {
                  element = DOMDocument.createElementNS('http://www.w3.org/2000/svg', name);
              } else {
                  element = DOMDocument.createElement(name);
                  element.dataset.id = elementId;
              }

              setAttributes(element, elementId, props, [], accessMap);
          }
          elementMap.set(elementId, element);

          injectElement(element, elementMap.get(parentId), index);
          if (!isUndefined(props._didInsert)) {
              postProcessing.push(function () {
                  return lifecycleEventFactory('_didInsert', props._didInsert);
              });
          }
      }

      function updateElement(patch) {
          var elementId = patch[PATCH_ELEMENT_ID];
          var updateAttrs = patch[PATCH_DATA];
          var rmAttrs = patch[PATCH_RM_DATA];
          var accessMap = patch[PATCH_AM_DATA];

          // $FlowIgnore
          setAttributes(elementMap.get(elementId), elementId, updateAttrs, rmAttrs, accessMap);
      }

      function removeElement(patch) {
          var elementId = patch[PATCH_ELEMENT_ID];
          if (elementId && elementMap.has(elementId)) {
              (function () {
                  // $FlowIgnore
                  var element = elementMap.get(elementId);
                  element.parentNode && element.parentNode.removeChild(element);
                  layoutRequests.delete(elementId);
                  layoutValues.delete(elementId);
                  elementMap.delete(elementId);
                  if (element.dataset && element.dataset._didRemove) {
                      postProcessing.push(function () {
                          return lifecycleEventFactory('_didRemove', element.dataset._didRemove);
                      });
                  }
              })();
          }
      }

      function registerStyle(patch) {
          var selector = patch[1];
          var css = patch[2];
          var element = styleMap.get(selector);
          if (!element) {
              element = DOMDocument.createElement('style');
              styleMap.set(selector, element);
              DOMDocument.head.appendChild(element);
          }
          if (element.textContent !== css) {
              if (!css) {
                  element.parentNode.removeChild(element);
                  styleMap.delete(selector);
              } else {
                  element.textContent = css;
              }
          }
      }

      function maintenance() {
          for (var key in styleMap.keys()) {
              if (!DOMDocument.querySelector('.' + key)) {
                  var element = styleMap.get(key);
                  styleMap.delete(key);
                  element.parentNode.removeChild(element);
              }
          }

          var _iteratorNormalCompletion3 = true;
          var _didIteratorError3 = false;
          var _iteratorError3 = undefined;

          try {
              var _loop2 = function _loop2() {
                  var elementId = _step3.value;

                  var id = layoutRequests.get(elementId);
                  var rect = elementMap.get(elementId).getBoundingClientRect();
                  var previousLayout = layoutValues.get(elementId) || {};

                  if (RECTANGLE_PROPS.some(function (key) {
                      return previousLayout[key] !== rect[key];
                  })) {
                      (function () {
                          var layoutEvt = {
                              id: id,
                              type: '_onLayout'
                          };
                          RECTANGLE_PROPS.forEach(function (key) {
                              return layoutEvt[key] = rect[key];
                          });

                          layoutValues.set(elementId, layoutEvt);
                          dispatchEvent(layoutEvt);
                      })();
                  };
              };

              for (var _iterator3 = layoutRequests.keys()[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                  _loop2();
              }
          } catch (err) {
              _didIteratorError3 = true;
              _iteratorError3 = err;
          } finally {
              try {
                  if (!_iteratorNormalCompletion3 && _iterator3.return) {
                      _iterator3.return();
                  }
              } finally {
                  if (_didIteratorError3) {
                      throw _iteratorError3;
                  }
              }
          }
      }

      var ACTION_METHODS = (_ACTION_METHODS = {}, defineProperty(_ACTION_METHODS, ADD_ACTION, addElement), defineProperty(_ACTION_METHODS, UPDATE_ACTION, updateElement), defineProperty(_ACTION_METHODS, REMOVE_ACTION, removeElement), defineProperty(_ACTION_METHODS, REGISTER_STYLE, registerStyle), _ACTION_METHODS);

      var pause = function pause() {
          paused = true;
      };

      var resume = function resume() {
          paused = false;
          processPatch();
      };

      var queuePatch = function queuePatch(patchSet) {
          var finish = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

          if (!patchSet) {
              return;
          }
          pendingPatches.unshift(patchSet.map(annotatePatchAttributes));
          if (finish) {
              processPatch();
              pause();
          } else {
              requestAnimationFrame(processPatch, domRoot);
          }
      };

      var processPatch = function processPatch() {
          if (paused) {
              return;
          }

          var patchSet = pendingPatches.pop();
          if (patchSet) {
              // console.group('PatchSet');
              var patch = void 0;
              while (patch = patchSet.shift()) {
                  // console.log(JSON.stringify(patch));
                  ACTION_METHODS[patch[PATCH_TYPE]](patch);
              }

              while (postProcessing.length) {
                  postProcessing.pop()();
              }
              // console.groupEnd('PatchSet');
          }
          if (pendingPatches.length) {
              requestAnimationFrame(processPatch, domRoot);
          } else {
              maintenance();
          }
      };

      requestAnimationFrame(processPatch, domRoot);

      return {
          queuePatch: queuePatch,
          pause: pause,
          resume: resume
      };
  }

  // Copyright (c) 2016 Timothy Farrell
  //
  // Licensed under the Apache License, Version 2.0 (the "License");
  // you may not use this file except in compliance with the License.
  // You may obtain a copy of the License at
  //
  //     http://www.apache.org/licenses/LICENSE-2.0
  //
  // Unless required by applicable law or agreed to in writing, software
  // distributed under the License is distributed on an "AS IS" BASIS,
  // WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  // See the License for the specific language governing permissions and
  // limitations under the License.


  function tryJSON(type, destination, id, data) {
      var packet = [type, destination, id, data];
      try {
          return JSON.stringify(packet);
      } catch (e) {
          return packet;
      }
  }

  function WorkerPortal(context, worker, isSlave, serialize) {
      var responseMap = new Map();
      var _worker = worker || self;
      var contextIndex = Object.keys(context);
      var _serialize = serialize ? function (type, destination, id, params) {
          return serialize(type, destination, id, params, tryJSON);
      } : tryJSON;
      var callCount = 0;
      var enabled = false;

      function post(type, id, destination, params) {
          _worker.postMessage(_serialize(type, destination, id, params));
      }

      function dispatcher(evt) {
          var data = void 0;
          try {
              data = JSON.parse(evt.data);
          } catch (e) {
              data = evt.data;
          }
          var destination = data[1];
          var id = data[2];
          var params = data[3];

          function _resolve(value) {
              post(1, id, 0, value);
          }
          function _reject(e) {
              post(1, id, 1, { message: e.message, name: e.name, stack: e.stack });
          }

          // If we have received an RPC response, satisfy the promise.
          if (data[0]) {
              if (responseMap.has(id)) {
                  var responses = responseMap.get(id);
                  responseMap.delete(id);
                  responses[destination](params);
              }
              return;
          }

          // If we have received an RPC call, execute and respond.
          var thennable = void 0;
          try {
              thennable = context[contextIndex[destination]].apply(null, params);
          } catch (e) {
              _reject(e);
          }
          if (thennable && typeof thennable.then === 'function') {
              thennable.then(_resolve).catch(_reject);
          } else {
              _resolve(thennable);
          }
      }

      function injectionPointFactory(fnId, callbackFactory) {
          return function () {
              for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
                  args[_key] = arguments[_key];
              }

              return new Promise(function (resolve, reject) {
                  if (!enabled && fnId !== 0) {
                      reject(new Error('Portal disabled'));
                  }
                  var id = callCount;
                  callCount += 1;
                  responseMap.set(id, [callbackFactory ? callbackFactory(resolve) : resolve, reject]);
                  post(0, id, fnId, args);
              });
          };
      }

      function resolveExternalInterfaceFactory(resolve) {
          return function (linkedFunctionNames) {
              var externalInterface = {};
              linkedFunctionNames.forEach(function (fnName, index) {
                  externalInterface[fnName] = injectionPointFactory(index);
              });
              enabled = true;
              resolve(externalInterface);
              return contextIndex;
          };
      }

      function cleanup() {
          enabled = false;
          _worker.removeEventListener('message', dispatcher);
          var _iteratorNormalCompletion = true;
          var _didIteratorError = false;
          var _iteratorError = undefined;

          try {
              for (var _iterator = responseMap.keys()[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                  var key = _step.value;

                  try {
                      responseMap.get(key)[1](new Error('Portal cleanup called.'));
                  } catch (e) {}
              }
          } catch (err) {
              _didIteratorError = true;
              _iteratorError = err;
          } finally {
              try {
                  if (!_iteratorNormalCompletion && _iterator.return) {
                      _iterator.return();
                  }
              } finally {
                  if (_didIteratorError) {
                      throw _iteratorError;
                  }
              }
          }
      }

      _worker.addEventListener('message', dispatcher);

      if (isSlave) {
          return new Promise(function (resolve) {
              contextIndex.splice(0, 0, '__init', '__cleanupSlave');
              context.__init = resolveExternalInterfaceFactory(resolve);
              context.__cleanupSlave = cleanup;
          });
      }

      return injectionPointFactory(0, resolveExternalInterfaceFactory)(contextIndex).then(function (api) {
          return Object.assign(api, {
              _cleanup: injectionPointFactory(1, function (resolve) {
                  return resolve(cleanup());
              })
          });
      });
  }

  var rootCounter = 0;

  function mount(componentOrWorker, domRoot, context) {
      // Worker doesn't exist for Chrome inside WebWorker context so we check for it.
      if (self.Worker && componentOrWorker instanceof self.Worker) {
          return mountWorker(componentOrWorker, domRoot);
      }
      return mountComponent(componentOrWorker, domRoot, context);
  }

  function mountComponent(component, domRoot, context) {
      var parent = allocateNode();
      var node = allocateNode();
      var instance = void 0;
      var rootId = rootCounter += 1;
      var rootContext = _extends({}, context);

      node.component = component;
      parent.key = getNodeKey(parent);
      node.key = getNodeKey(node, parent, 0);

      if (!domRoot) {
          return new Promise(function (resolve) {
              var portal = void 0;

              WorkerPortal({
                  mount: function mount(rootId) {
                      parent.id = rootId;
                      instance = Component(node, function (patches) {
                          portal.onpatch(patches);
                      }, parent, _extends({}, rootContext, portal));
                      resolve(instance);
                  },
                  dispatchEvent: function dispatchEvent(evt) {
                      return instance.dispatchEvent(evt);
                  },
                  update: function update(props) {
                      return instance.update(props);
                  },
                  remove: function remove() {
                      return instance.remove();
                  }
              }).then(function (p) {
                  portal = p;
              });
          });
      } else {
          var eventHandler = function eventHandler(evt) {
              instance.dispatchEvent(evt);
          };

          parent.id = rootId;

          instance = Component(node, Renderer(domRoot, rootId, eventHandler).queuePatch, parent, rootContext);
          return instance;
      }
  }

  function mountWorker(worker, domRoot, context) {
      var renderer = void 0;
      return WorkerPortal(_extends({
          onpatch: function onpatch(data) {
              renderer.queuePatch(data);
          }
      }, context), worker, true, function (type, destination, id, params, defaultSerializer) {
          try {
              return 'dataTransfer' in params[0] ? [type, destination, id, params] : defaultSerializer(type, destination, id, params);
          } catch (e) {
              return defaultSerializer(type, destination, id, params);
          }
      }).then(function (portal) {
          var rootId = rootCounter += 1;

          renderer = Renderer(domRoot, rootId, portal.dispatchEvent);
          portal.mount(rootId);

          return portal;
      }).catch(function (e) {
          console.error(e.stack);
      });
  }

  var toggleWeakMap = new WeakMap();

  function toggle(prop) {
      if (!toggleWeakMap.has(prop)) {
          toggleWeakMap.set(prop, function _toggle() {
              prop(!prop());
          });
      }

      return toggleWeakMap.get(prop);
  }

  function prop(store, callback) {
      var version = 0;

      var accessor = function _prop() {
          var args = asArray(arguments);
          if (args.length) {
              version += 1;
              store = args[0];
              if (callback) {
                  callback(store);
              }
          }
          return store;
      };
      accessor.v = function () {
          return version;
      };

      return accessor;
  }

  function computed(dependencies, fn) {
      var val = undefined;
      var version = 0;
      if (isFunction(dependencies)) {
          fn = dependencies;
          dependencies = [];
      }
      var previousVersions = [];

      var accessor = function _computed() {
          var currentVersions = dependencies.map(function (d) {
              return d.v();
          });

          if (currentVersions.some(function (v, i) {
              return v !== previousVersions[i];
          }) || !dependencies.length) {
              previousVersions = currentVersions;
              version++;
              val = fn.apply(null, dependencies.map(function (d) {
                  return d();
              }));
          }
          return val;
      };
      accessor.v = function () {
          accessor();
          return version;
      };

      return accessor;
  }

  function throttle(method) {
      var afterTime = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 16;

      var id = Null;
      function throttleInterval() {
          id = Null;
      }

      return function throttleWrapper() {
          if (id === Null) {
              id = setTimeout(throttleInterval, afterTime);
              method.apply(Null, arguments);
          }
      };
  }

  function debounce(method) {
      var afterTime = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 16;

      var id = void 0;
      var args = void 0;
      function clear() {
          if (id !== Null) {
              clearTimeout(id);
              id = Null;
          }
      }

      function debounceInterval() {
          clear();
          try {
              method.apply(Null, args);
          } catch (e) {
              args = [];
              throw e;
          }
      }

      var wrapped = function debounceWrapper() {
          id = setTimeout(debounceInterval, afterTime);
          args = asArray(arguments);
      };

      wrapped.clear = clear;

      return wrapped;
  }



  var _addons = Object.freeze({
  	toggle: toggle,
  	prop: prop,
  	computed: computed,
  	memoize: memoize,
  	throttle: throttle,
  	debounce: debounce
  });

  exports.addons = _addons;
  exports.Router = Router;
  exports.Component = Component;
  exports.Renderer = Renderer;
  exports.mount = mount;
  exports.mountWorker = mountWorker;

  Object.defineProperty(exports, '__esModule', { value: true });

}));