'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.mount = mount;
exports.mountComponent = mountComponent;
exports.mountWorker = mountWorker;

var _component = require('./component');

var _node = require('./node');

var _renderer = require('./renderer');

var _workerPortal = require('worker-portal');

var rootCounter = 0;

function mount(componentOrWorker, domRoot, context) {
    // Worker doesn't exist for Chrome inside WebWorker context so we check for it.
    if (self.Worker && componentOrWorker instanceof self.Worker) {
        return mountWorker(componentOrWorker, domRoot);
    }
    return mountComponent(componentOrWorker, domRoot, context);
}

function mountComponent(component, domRoot, context) {
    var parent = (0, _node.allocateNode)();
    var node = (0, _node.allocateNode)();
    var instance = void 0;
    var rootId = rootCounter += 1;
    var rootContext = _extends({}, context);

    node.component = component;
    parent.key = (0, _node.getNodeKey)(parent);
    node.key = (0, _node.getNodeKey)(node, parent, 0);

    if (!domRoot) {
        return new Promise(function (resolve) {
            var portal = void 0;

            (0, _workerPortal.WorkerPortal)({
                mount: function mount(rootId) {
                    parent.id = rootId;
                    instance = (0, _component.Component)(node, function (patches) {
                        portal.onpatch(patches);
                    }, parent, _extends({}, rootContext, portal));
                    resolve(instance);
                },
                dispatchEvent: function dispatchEvent(evt) {
                    return instance.dispatchEvent(evt);
                },
                update: function update(props) {
                    return instance.update(props);
                },
                remove: function remove() {
                    return instance.remove();
                }
            }).then(function (p) {
                portal = p;
            });
        });
    } else {
        var eventHandler = function eventHandler(evt) {
            instance.dispatchEvent(evt);
        };

        parent.id = rootId;

        instance = (0, _component.Component)(node, (0, _renderer.Renderer)(domRoot, rootId, eventHandler).queuePatch, parent, rootContext);
        return instance;
    }
}

function mountWorker(worker, domRoot, context) {
    var renderer = void 0;
    return (0, _workerPortal.WorkerPortal)(_extends({
        onpatch: function onpatch(data) {
            renderer.queuePatch(data);
        }
    }, context), worker, true, function (type, destination, id, params, defaultSerializer) {
        try {
            return 'dataTransfer' in params[0] ? [type, destination, id, params] : defaultSerializer(type, destination, id, params);
        } catch (e) {
            return defaultSerializer(type, destination, id, params);
        }
    }).then(function (portal) {
        var rootId = rootCounter += 1;

        renderer = (0, _renderer.Renderer)(domRoot, rootId, portal.dispatchEvent);
        portal.mount(rootId);

        return portal;
    }).catch(function (e) {
        console.error(e.stack);
    });
}