'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.Component = Component;

var _node2 = require('./node');

var _utils = require('./utils');

var _trimkit = require('trimkit');

var _constants = require('./constants');

var eventCounter = 0;
var existingEventMap = new Map();
var events = new Map();
var selectorMatcher = /(?:(?:}|(?:(?:@[^{]+){))\s*)([^,@{]+)(?:,\s*([^,@{]+))*/gm;

function Component(node, onpatch, parent) {
    var parentContext = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};

    var state = [];
    var rawState = [];
    var pendingPatches = [];
    var checkWillUpdate = _trimkit.Null;
    var updateLock = false;
    var updateRequested = false;
    var instance = void 0;
    var previousProps = void 0;
    var previousChildren = [];
    var parentOffset = 0;
    var version = 0;
    // Clone the nodes so they can be freed in the parent's diff process.
    var hostNode = _extends({}, node);
    var parentNode = _extends({}, parent);
    var context = _extends({}, parentContext);
    var componentMap = new Map();

    function dispatchEvent(data) {
        var callback = events.get(data.id);
        if (callback) {
            // Flow doesn't like my PseudoEventType going into a real EventHandler. Can't imagine why.
            // $FlowIgnore
            return callback(data);
        }
    }

    function replaceEventHandlers(_node) {
        var name = _node.name,
            attrs = _node.attrs,
            id = _node.id;

        for (var key in attrs) {
            var value = attrs[key];
            if (key.startsWith('on') || key[0] === '_' && (0, _trimkit.isFunction)(value)) {
                if (!existingEventMap.has(id)) {
                    existingEventMap.set(id, {});
                }
                var elementRecord = existingEventMap.get(id);
                var eventId = elementRecord[key];

                if ((0, _trimkit.isUndefined)(eventId)) {
                    eventId = eventCounter++;
                    elementRecord[key] = eventId;
                    events.set(eventId, value);
                } else if (value !== events.get(eventId)) {
                    events.set(eventId, value);
                }
                var flagKey = '_' + key + '_flags';
                var flags = attrs[flagKey];
                attrs[key] = eventId;
                if (flags) {
                    attrs[flagKey] = flags.join(',');
                }
            }
        }

        return attrs;
    }

    function removeEventHandlers(elementId) {
        var elementRecord = existingEventMap.get(elementId);
        if ((0, _trimkit.isUndefined)(elementRecord)) {
            return;
        }

        for (var eventName in elementRecord) {
            events.delete(elementRecord[eventName]);
        }
        existingEventMap.delete(elementId);
    }

    function addUpdateChildComponent(parent, _node, indexOffset) {
        var key = _node.key;

        var childInstance = componentMap.get(key);
        if (!childInstance) {
            childInstance = Component(_node, onpatch, parent, context);
            componentMap.set(key, childInstance);
        }

        var _childInstance$getPat = childInstance.getPatches(_node.attrs, _node.children, indexOffset),
            _childInstance$getPat2 = _slicedToArray(_childInstance$getPat, 3),
            patches = _childInstance$getPat2[0],
            childRequestsUpdate = _childInstance$getPat2[1],
            childCount = _childInstance$getPat2[2];

        updateRequested = updateRequested || childRequestsUpdate;
        return [patches, childCount];
    }

    function addNode(_node, parent, index) {
        if (_node.component) {
            return addUpdateChildComponent(parent, _node, index);
        }
        var name = _node.name,
            component = _node.component;

        var patchAttrs = _node.attrs ? replaceEventHandlers(_node) : {};
        if (patchAttrs._willInsert) {
            updateLock = true;
            dispatchEvent({
                type: '_willInsert',
                id: patchAttrs._willInsert
            });
            updateLock = false;
        }
        return [[[_constants.ADD_ACTION, _node.id, [component ? component.name : name, patchAttrs, parent.id, index]]].concat(_node.static ? [] : diff([], _node.children, _node, 0)), 1];
    }

    function removeNode(_node) {
        var patches = void 0;
        var processRemove = function processRemove() {
            removeEventHandlers(_node.id);
            return diff(_node.children, [], _node).concat([[_constants.REMOVE_ACTION, _node.id]]);
        };

        if (_node.component) {
            var componentInstance = componentMap.get(_node.key);
            if (componentInstance) {
                patches = componentInstance.remove();
                componentMap.delete(_node.key);
            }
        } else if (_node.attrs._willRemove) {
            updateLock = true;
            var thennable = dispatchEvent({
                type: '_willRemove',
                id: _node.attrs._willRemove
            });
            updateLock = false;
            if (thennable && (0, _trimkit.isFunction)(thennable.then)) {
                _node.delayFree = true;
                thennable.then(function () {
                    onpatch(processRemove());
                    _node.delayFree = false;
                    (0, _node2.freeNode)(_node);
                });
                patches = [];
            } else {
                patches = processRemove();
            }
        } else {
            patches = processRemove();
        }
        return patches;
    }

    function diff(currentState, nextState, parent, indexOffset) {
        var patches = [];
        var currentCount = currentState.length;
        var nextCount = nextState.length;

        var index = 0;

        while (index < currentCount) {
            var nextNode = index < nextCount ? nextState[index] : _trimkit.Null;
            var current = currentState[index];

            if (!nextNode) {
                patches.push.apply(patches, removeNode(current));
                if (current.delayFree) {
                    nextState[index] = current;
                    nextNode = current;
                }
            } else if ((0, _node2.shouldReplaceNode)(current, nextNode)) {
                patches.push.apply(patches, removeNode(current));
                if (current.delayFree) {
                    nextState[index] = current;
                    (0, _node2.freeNode)(nextNode);
                    nextNode = current;
                } else {
                    var _addNode = addNode(nextNode, parent, index + indexOffset),
                        _addNode2 = _slicedToArray(_addNode, 2),
                        nodePatches = _addNode2[0],
                        childCount = _addNode2[1];

                    patches.push.apply(patches, nodePatches);
                    index += childCount - 1;
                }
            } else if (nextNode.component) {
                // nodes are roughly equal but they're components so we compare them differently
                nextNode.key = current.key;

                var _addUpdateChildCompon = addUpdateChildComponent(parent, nextNode, index + indexOffset),
                    _addUpdateChildCompon2 = _slicedToArray(_addUpdateChildCompon, 2),
                    _nodePatches = _addUpdateChildCompon2[0],
                    _childCount = _addUpdateChildCompon2[1];

                patches.push.apply(patches, _nodePatches);
                index += _childCount - 1;
            } else {
                nextNode.id = current.id;

                var currentAttrs = current.attrs;
                var normalizedNextAttributes = nextNode.attrs ? replaceEventHandlers(nextNode) : {};
                var newAttrs = {};
                var rmAttrs = [];
                var attrsAreEqual = true;

                for (var key in currentAttrs) {
                    var val = normalizedNextAttributes[key];
                    if (!(0, _trimkit.isUndefined)(val)) {
                        if (val !== currentAttrs[key]) {
                            newAttrs[key] = val;
                            attrsAreEqual = false;
                        }
                    } else {
                        rmAttrs.push(key);
                        attrsAreEqual = false;
                    }
                }

                for (var _key in normalizedNextAttributes) {
                    if ((0, _trimkit.isUndefined)(currentAttrs[_key])) {
                        newAttrs[_key] = normalizedNextAttributes[_key];
                        attrsAreEqual = false;
                    }
                }

                if (!attrsAreEqual) {
                    // Update the current element
                    patches.push([_constants.UPDATE_ACTION, current.id, newAttrs, rmAttrs]);
                }

                if (!nextNode.static) {
                    patches.push.apply(patches, diff(current.children, nextNode.children, nextNode, 0));
                }
            }

            index++;
        }

        while (index < nextCount) {
            var _nextNode = nextState[index];

            var _addNode3 = addNode(_nextNode, parent, index + indexOffset),
                _addNode4 = _slicedToArray(_addNode3, 2),
                _nodePatches2 = _addNode4[0],
                _childCount2 = _addNode4[1];

            patches.push.apply(patches, _nodePatches2);
            index += _childCount2; // - 1 + 1  -1 to correct for the count, +1 for adding this node.
        }

        return patches;
    }

    function updateChildContext(c) {
        (0, _trimkit.ObjectKeys)(c).forEach(function (k) {
            return context[k] = c[k];
        });
    }

    function registerStyle(css) {
        var key = getKey();
        hostNode.style.push(key);
        var outCSS = ('} ' + css.trim()).replace(selectorMatcher, function (match) {
            for (var _len = arguments.length, p = Array(_len > 1 ? _len - 1 : 0), _key2 = 1; _key2 < _len; _key2++) {
                p[_key2 - 1] = arguments[_key2];
            }

            p.slice(0, p.length - 2).forEach(function (s) {
                if (s) {
                    match = match.split(s).join(s.trim() + '[' + key + ']');
                }
            });
            return match;
        }).slice(2);

        pendingPatches.push([_constants.REGISTER_STYLE, getKey(), outCSS]);
    }

    function getKey() {
        return hostNode.key;
    }

    function handleUpdate(checkFunction) {
        checkWillUpdate = checkFunction;
    }

    function update(props) {
        if (updateLock) {
            updateRequested = true;
            return;
        }

        var _getPatches2 = getPatches(props),
            _getPatches3 = _slicedToArray(_getPatches2, 2),
            patches = _getPatches3[0],
            childRequestsUpdate = _getPatches3[1];

        patches.length && onpatch(patches);
        if (childRequestsUpdate) {
            redraw();
        }
    }

    function _getPatches() {
        var newState = (0, _node2.convertTagsToNodes)(hostNode, rawState);
        var patches = diff(state, newState, parentNode, parentOffset);
        if (pendingPatches.length) {
            patches.push.apply(patches, pendingPatches);
            pendingPatches = [];
        }
        state.map(_node2.freeNode);
        state = newState;
        return patches;
    }

    function versionedRedrawFactory(frozenVersion) {
        return function versionedLockout() {
            if (frozenVersion === version) {
                redraw();
            }
        };
    }

    function getPatches(props, children, indexOffset) {
        if ((0, _trimkit.isUndefined)(props)) {
            props = previousProps;
        }

        if ((0, _trimkit.isUndefined)(children)) {
            children = previousChildren;
        } else {
            previousChildren = children;
        }

        if (!(0, _trimkit.isUndefined)(indexOffset)) {
            parentOffset = indexOffset;
        }

        updateLock = true;
        if (instance && checkWillUpdate && !checkWillUpdate(props, previousProps)) {
            previousProps = props;
        } else if (!instance || !(0, _utils.shallowEqual)(previousProps, props)) {
            version += 1;
            instance = hostNode.component(_extends({
                redraw: versionedRedrawFactory(version),
                getKey: getKey,
                registerStyle: registerStyle,
                handleUpdate: handleUpdate,
                updateChildContext: updateChildContext
            }, context), props);
            previousProps = props;
        }
        updateLock = false;

        rawState = instance(children);
        var patches = _getPatches();
        var isRequestingUpdate = updateRequested;
        updateRequested = false;
        var childCount = (0, _utils.isTag)(rawState) ? 1 : rawState.length;
        return [patches, isRequestingUpdate, childCount];
    }

    function redraw() {
        if (!instance) {
            return;
        }
        update();
    }

    function remove() {
        version += 1;
        rawState = [];
        return _getPatches();
    }

    return {
        remove: remove,
        redraw: redraw,
        update: update,
        getPatches: getPatches,
        dispatchEvent: dispatchEvent
    };
}