'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

exports.allocateNode = allocateNode;
exports.freeNode = freeNode;
exports.isObject = isObject;
exports.isComponent = isComponent;
exports.shouldReplaceNode = shouldReplaceNode;
exports.processTagClass = processTagClass;
exports.parseTagAttrs = parseTagAttrs;
exports.getNodeKey = getNodeKey;
exports.convertTagToNodes = convertTagToNodes;
exports.convertTagsToNodes = convertTagsToNodes;

var _constants = require('./constants');

var _trimkit = require('trimkit');

var _utils = require('./utils');

var pool = [];
var nodeCounter = 0;

function allocateNode() {
    var node = pool.length > 0 ? pool.pop() : {
        id: _trimkit.Null,
        name: '',
        attrs: {},
        children: [],
        static: false,
        svg: false,
        component: _trimkit.Null,
        key: '',
        style: [],
        delayFree: false
    };
    node.id = ++nodeCounter;
    return node;
}

function freeNode(node) {
    if (node.delayFree) {
        return;
    }
    if (!node.component) {
        node.children.map(freeNode);
    }
    node.component = _trimkit.Null;
    node.static = node.svg = false;
    node.attrs = {};
    node.name = node.key = '';
    node.children = node.style = [];

    pool.push(node);
}

function isObject(obj) {
    return (typeof obj === 'undefined' ? 'undefined' : _typeof(obj)) === 'object' && _trimkit.Null !== obj && !(0, _trimkit.isArray)(obj);
}

function isComponent(obj) {
    return (0, _trimkit.isArray)(obj) && (0, _trimkit.isFunction)(obj[_constants.TAG_NAME]);
}

function shouldReplaceNode(currentNode, nextNode) {
    return nextNode.name !== currentNode.name || nextNode.component !== currentNode.component;
}

function processTagClass(classAttr) {
    var classObj = (0, _utils.traverseThunk)(classAttr);
    if (isObject(classObj)) {
        classObj = (0, _trimkit.ObjectKeys)(classObj).filter(function (key) {
            return (0, _utils.traverseThunk)(classObj[key]);
        });
    } else if ((0, _trimkit.isString)(classObj)) {
        classObj = classObj.split(' ');
    }
    return classObj;
}

var attributesParser = /(?:(^|#|\.)([^#\.\[\]]+))|(\[(.+?)(?:\s*=\s*("|'|)((?:\\["'\]]|.)*?)\5)?\])/g;

function parseTagAttrs(signature, attrs) {
    var _parseTagSignature = parseTagSignature(signature),
        _parseTagSignature2 = _slicedToArray(_parseTagSignature, 3),
        tag = _parseTagSignature2[0],
        sigAttrs = _parseTagSignature2[1],
        classes = _parseTagSignature2[2];

    var newAttrs = _extends({}, sigAttrs, {
        class: classes.concat(processTagClass(attrs['class'] || '')).concat(processTagClass(attrs.className || '')).join(' ').trim()
    });

    (0, _trimkit.ObjectKeys)(attrs).forEach(function (key) {
        if (key.indexOf('class') === 0) {
            return;
        }
        var value = attrs[key];
        newAttrs[key] = key.indexOf('on') === 0 || key[0] === '_' ? value : (0, _utils.traverseThunk)(value);
    });

    return [tag, newAttrs];
};

var parseTagSignature = (0, _utils.memoize)(function (signature) {
    var tag = 'div';
    var classes = [];
    var match = void 0;
    var attrs = {};

    while (match = attributesParser.exec(signature)) {
        if (match[1] === '' && match[2]) {
            tag = match[2];
        } else if (match[1] === '#') {
            attrs.id = match[2];
        } else if (match[1] === '.') {
            classes.push(match[2]);
        } else if (match[3][0] === '[') {
            var attrValue = match[6];
            if (attrValue) attrValue = attrValue.replace(/\\(["'])/g, '$1');
            attrs[match[4]] = attrValue || true;
        }
    }

    return [tag, attrs, classes];
});

function getNodeKey(node, parent, index) {
    return node.attrs._key ? (node.component.name || 'tag') + '-' + node.attrs._key : parent ? 'sl_' + parent.id.toString(36) + '_' + index.toString(36) : 'sl_root';
}

function convertTagToNodes(parent, tag, index) {
    // tag can be a thunk, string, array of tags or array of arrays.
    // we need to deal with each.

    // Thunk

    // FIXME - There could be an issue here where a thunk and a component could
    // be confused if the thunk is the first thing out of a render function.
    // Lock this down!
    if ((0, _trimkit.isFunction)(tag)) {
        var state = (0, _utils.traverseThunk)(tag, _constants.GENERIC_ERROR_MESSAGE);
        return convertTagsToNodes(parent, state);
    }

    var node = allocateNode();

    // String (Text tag)
    if ((0, _trimkit.isString)(tag) || (0, _trimkit.isNumber)(tag)) {
        node.name = _constants.TEXT_NODE_VALUE;
        node.attrs = {
            textContent: tag.toString()
        };
        node.static = true;
    } else if ((0, _utils.isTag)(tag)) {
        // Element tag

        var hasAttrs = isObject(tag[_constants.TAG_ATTRS]);
        var attrs = hasAttrs ? tag[_constants.TAG_ATTRS] : {};

        var _parseTagAttrs = parseTagAttrs(tag[_constants.TAG_NAME], attrs);

        var _parseTagAttrs2 = _slicedToArray(_parseTagAttrs, 2);

        node.name = _parseTagAttrs2[0];
        node.attrs = _parseTagAttrs2[1];

        node.key = node.attrs._key || node.attrs.id || _trimkit.Null;

        if (parent.svg || node.name === 'svg') {
            node.svg = true;
            node.attrs._ns = 'svg';
        }
        node.style = parent.style;
        node.style.forEach(function (s) {
            node.attrs[s] = '';
        }); // Add style attributes to the tag.
        node.children = convertTagChildrenToNodes(node, tag.slice(hasAttrs ? _constants.TAG_CHILDREN : _constants.TAG_ATTRS));
    } else if (isComponent(tag)) {
        var _hasAttrs = isObject(tag[_constants.TAG_ATTRS]);
        var _attrs = _hasAttrs ? tag[_constants.TAG_ATTRS] : {};

        node.component = tag[_constants.TAG_NAME];
        node.attrs = _attrs;
        node.key = getNodeKey(node, parent, index);
        node.style = parent.style.slice();
        node.static = true;
        node.children = tag.slice(_hasAttrs ? _constants.TAG_CHILDREN : _constants.TAG_ATTRS);
    } else {
        freeNode(node);
        // Array
        return convertTagsToNodes(parent, tag);
    }

    return node;
}

function flatten(arr) {
    return arr.reduce(function (flat, toFlatten) {
        return flat.concat((0, _trimkit.isArray)(toFlatten) && !(0, _utils.isTag)(toFlatten) ? flatten(toFlatten) : toFlatten);
    }, []);
}

function convertTagChildrenToNodes(parent, tags) {
    if ((0, _trimkit.isUndefined)(tags) || _trimkit.Null === tags) {
        return [];
    }
    return flatten(tags.map(function (t, i) {
        return convertTagToNodes(parent, t, i);
    }));
}

function convertTagsToNodes(parent, tags) {
    if ((0, _utils.isTag)(tags) || (0, _trimkit.isString)(tags) || isComponent(tags)) {
        return [convertTagToNodes(parent, tags, 0)];
    }
    return convertTagChildrenToNodes(parent, tags);
}