'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

exports.annotatePatchAttributes = annotatePatchAttributes;
exports.annotateAttribute = annotateAttribute;
exports.Renderer = Renderer;

var _constants = require('./constants');

var _trimkit = require('trimkit');

var _utils = require('./utils');

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var RECTANGLE_PROPS = 'height width top right bottom left x y'.split(' ');

function injectElement(element, parent, index) {
    var nextSibling = parent.childNodes[index];
    if (nextSibling) {
        parent.insertBefore(element, nextSibling);
    } else {
        parent.appendChild(element);
    }
}

function callPropertyFactory(element, propName) {
    return function () {
        return element[propName]();
    };
}

function annotatePatchAttributes(patch) {
    if (patch[_constants.PATCH_TYPE] !== _constants.ADD_ACTION && patch[_constants.PATCH_TYPE] !== _constants.UPDATE_ACTION) {
        return patch;
    }

    var accessMap = {};
    var patchData = patch[_constants.PATCH_DATA];
    var patchAttrs = patch[_constants.PATCH_TYPE] === _constants.ADD_ACTION ? patchData[1] : patchData;
    for (var name in patchAttrs) {
        accessMap[name] = annotateAttribute(name);
    }

    if (patch[_constants.PATCH_TYPE] === _constants.UPDATE_ACTION) {
        patch[_constants.PATCH_RM_DATA].forEach(function (name) {
            accessMap[name] = annotateAttribute(name);
        });
    }

    patch.push(accessMap);

    return patch;
}

var ACTION_ATTRIBUTES = ['focus', 'select', 'blur', 'click'];
var ELEMENT_PROPERTIES = 'textContent value checked disabled readOnly scrollTop scrollLeft'.split(' ');
var LOCAL_LIFECYCLE_PROPERTIES = ['_didInsert', '_willInsert', '_willRemove'];
var LIFECYCLE_PROPERTIES = '_didRemove';
var LAYOUT_REQUEST = '_onLayout';
function annotateAttribute(name) {
    if (ELEMENT_PROPERTIES.indexOf(name) !== -1) {
        return _constants.ELEMENT_PROPERTY;
    } else if (LOCAL_LIFECYCLE_PROPERTIES.indexOf(name) !== -1) {
        return _constants.LOCAL_ELEMENT_LIFECYCLE_EVENT;
    } else if (LIFECYCLE_PROPERTIES === name) {
        return _constants.ELEMENT_LIFECYCLE_EVENT;
    } else if (LAYOUT_REQUEST === name) {
        return _constants.LAYOUT_REQUEST_EVENT;
    } else if (ACTION_ATTRIBUTES.indexOf(name) !== -1) {
        return _constants.RUN_AFTER_ACTION;
    } else if (name.indexOf('on') === 0) {
        return _constants.ELEMENT_EVENT;
    }
    return _constants.ELEMENT_ATTRIBUTE;
}

function Renderer(domRoot) {
    var _ACTION_METHODS;

    var rootId = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
    var dispatchEvent = arguments[2];

    var elementMap = new Map();
    var eventMap = new Map();
    var styleMap = new Map();
    var pendingPatches = [];
    var paused = false;
    var postProcessing = [];
    var layoutRequests = new Map();
    var layoutValues = new Map();

    window.addEventListener('resize', maintenance, false);

    elementMap.set(rootId, domRoot);

    function lifecycleEventFactory(type, id) {
        (0, _trimkit.requestAnimationFrame)(function () {
            dispatchEvent({
                type: type,
                id: parseInt(id, 10)
            });
        });
    }

    function eventHandler(evt) {
        if (!evt instanceof Event) {
            // Lifecycle events are not DOM events.
            dispatchEvent(evt);
        }
        var eventName = evt.type;
        var dataset = evt.currentTarget.dataset;
        var id = parseInt(dataset['on' + eventName + 'EventId'], 10);
        var flags = eventMap.get(id).split(',');
        var fakeEvt = {
            id: id,
            type: eventName
        };
        var stop = true;
        var index = flags.length;

        var _loop = function _loop() {
            var flag = flags[index];
            if (flag === 'bubble') {
                stop = false;
                return 'continue';
            }
            var flagPaths = flag.split('.');
            var pathLastIndex = flagPaths.length - 1;
            var pointer = evt;
            var fakeEvtPointer = fakeEvt;
            flagPaths.forEach(function (attr, i) {
                var newPointer = pointer[attr];
                if (i !== pathLastIndex) {
                    pointer = newPointer;
                    fakeEvtPointer = fakeEvtPointer[attr] || (fakeEvtPointer[attr] = {});
                } else if ((0, _trimkit.isFunction)(newPointer)) {
                    pointer[attr](); // run with context
                } else {
                    fakeEvtPointer[attr] = newPointer;
                }
            });
        };

        while (index--) {
            var _ret = _loop();

            if (_ret === 'continue') continue;
        }
        if (stop) {
            evt.stopPropagation();
        }
        dispatchEvent(fakeEvt);
    }

    function setAttributes(element, elementId, attrs, rmAttrs, accessMap) {
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
            for (var _iterator = (0, _trimkit.ObjectKeys)(attrs)[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                var name = _step.value;

                var _value = attrs[name];

                if (_value === _trimkit.Null) {
                    continue;
                }

                switch (accessMap[name]) {
                    case _constants.ELEMENT_PROPERTY:
                        {
                            element[name] = _value;
                            continue;
                        }
                    case _constants.ELEMENT_EVENT:
                        {
                            var _eventName = name.slice(2);
                            var eventIdAttrName = name + 'EventId';
                            if (element.dataset[eventIdAttrName]) {
                                element.removeEventListener(_eventName, eventHandler);
                            }
                            if (_value) {
                                element.dataset[eventIdAttrName] = _value;
                                eventMap.set(_value, attrs['_' + name + '_flags'] || (0, _utils.getEventFlags)(_eventName, element.tagName.toLowerCase(), attrs.type));
                                element.addEventListener(_eventName, eventHandler, false);
                            }
                            continue;
                        }
                    case _constants.ELEMENT_LIFECYCLE_EVENT:
                        {
                            element.dataset[name] = _value;
                            continue;
                        }
                    case _constants.ELEMENT_ATTRIBUTE:
                        {
                            element.setAttribute(name, _value);
                            continue;
                        }
                    case _constants.LAYOUT_REQUEST_EVENT:
                        {
                            layoutRequests.set(elementId, _value);
                            continue;
                        }
                    case _constants.RUN_AFTER_ACTION:
                        {
                            if (_value) {
                                postProcessing.push(callPropertyFactory(element, name));
                            }
                        }
                    default:
                }
            }
        } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion && _iterator.return) {
                    _iterator.return();
                }
            } finally {
                if (_didIteratorError) {
                    throw _iteratorError;
                }
            }
        }

        var _iteratorNormalCompletion2 = true;
        var _didIteratorError2 = false;
        var _iteratorError2 = undefined;

        try {
            for (var _iterator2 = rmAttrs[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                var data = _step2.value;

                var _data = _slicedToArray(data, 2),
                    accessCode = _data[0],
                    _name = _data[1];

                switch (accessCode) {
                    case _constants.ELEMENT_EVENT:
                        element.removeEventListener(eventName, eventHandler);
                        continue;
                    case _constants.ELEMENT_PROPERTY:
                        {
                            element[_name] = value;
                            continue;
                        }
                    case _constants.ELEMENT_ATTRIBUTE:
                        {
                            element.removeAttribute(_name, value);
                        }
                    case _constants.LAYOUT_REQUEST_EVENT:
                        {
                            layoutRequests.delete(elementId);
                            layoutValues.delete(elementId);
                            continue;
                        }

                    case _constants.ELEMENT_LIFECYCLE_EVENT:
                        {
                            element.dataset.removeProp(_name);
                            continue;
                        }
                    default:
                }
            }
        } catch (err) {
            _didIteratorError2 = true;
            _iteratorError2 = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion2 && _iterator2.return) {
                    _iterator2.return();
                }
            } finally {
                if (_didIteratorError2) {
                    throw _iteratorError2;
                }
            }
        }
    }

    function addElement(patch) {
        var elementId = patch[_constants.PATCH_ELEMENT_ID];
        var data = patch[_constants.PATCH_DATA];
        var accessMap = patch[_constants.PATCH_RM_DATA];

        var _data2 = _slicedToArray(data, 4),
            name = _data2[0],
            props = _data2[1],
            parentId = _data2[2],
            index = _data2[3];

        var element = void 0;

        if (name === _constants.TEXT_NODE_VALUE) {
            // $FlowIgnore
            element = _trimkit.DOMDocument.createTextNode(props.textContent);
        } else {
            if (props._ns === 'svg') {
                element = _trimkit.DOMDocument.createElementNS('http://www.w3.org/2000/svg', name);
            } else {
                element = _trimkit.DOMDocument.createElement(name);
                element.dataset.id = elementId;
            }

            setAttributes(element, elementId, props, [], accessMap);
        }
        elementMap.set(elementId, element);

        injectElement(element, elementMap.get(parentId), index);
        if (!(0, _trimkit.isUndefined)(props._didInsert)) {
            postProcessing.push(function () {
                return lifecycleEventFactory('_didInsert', props._didInsert);
            });
        }
    }

    function updateElement(patch) {
        var elementId = patch[_constants.PATCH_ELEMENT_ID];
        var updateAttrs = patch[_constants.PATCH_DATA];
        var rmAttrs = patch[_constants.PATCH_RM_DATA];
        var accessMap = patch[_constants.PATCH_AM_DATA];

        // $FlowIgnore
        setAttributes(elementMap.get(elementId), elementId, updateAttrs, rmAttrs, accessMap);
    }

    function removeElement(patch) {
        var elementId = patch[_constants.PATCH_ELEMENT_ID];
        if (elementId && elementMap.has(elementId)) {
            (function () {
                // $FlowIgnore
                var element = elementMap.get(elementId);
                element.parentNode && element.parentNode.removeChild(element);
                layoutRequests.delete(elementId);
                layoutValues.delete(elementId);
                elementMap.delete(elementId);
                if (element.dataset && element.dataset._didRemove) {
                    postProcessing.push(function () {
                        return lifecycleEventFactory('_didRemove', element.dataset._didRemove);
                    });
                }
            })();
        }
    }

    function registerStyle(patch) {
        var selector = patch[1];
        var css = patch[2];
        var element = styleMap.get(selector);
        if (!element) {
            element = _trimkit.DOMDocument.createElement('style');
            styleMap.set(selector, element);
            _trimkit.DOMDocument.head.appendChild(element);
        }
        if (element.textContent !== css) {
            if (!css) {
                element.parentNode.removeChild(element);
                styleMap.delete(selector);
            } else {
                element.textContent = css;
            }
        }
    }

    function maintenance() {
        for (var key in styleMap.keys()) {
            if (!_trimkit.DOMDocument.querySelector('.' + key)) {
                var element = styleMap.get(key);
                styleMap.delete(key);
                element.parentNode.removeChild(element);
            }
        }

        var _iteratorNormalCompletion3 = true;
        var _didIteratorError3 = false;
        var _iteratorError3 = undefined;

        try {
            var _loop2 = function _loop2() {
                var elementId = _step3.value;

                var id = layoutRequests.get(elementId);
                var rect = elementMap.get(elementId).getBoundingClientRect();
                var previousLayout = layoutValues.get(elementId) || {};

                if (RECTANGLE_PROPS.some(function (key) {
                    return previousLayout[key] !== rect[key];
                })) {
                    (function () {
                        var layoutEvt = {
                            id: id,
                            type: '_onLayout'
                        };
                        RECTANGLE_PROPS.forEach(function (key) {
                            return layoutEvt[key] = rect[key];
                        });

                        layoutValues.set(elementId, layoutEvt);
                        dispatchEvent(layoutEvt);
                    })();
                };
            };

            for (var _iterator3 = layoutRequests.keys()[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                _loop2();
            }
        } catch (err) {
            _didIteratorError3 = true;
            _iteratorError3 = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion3 && _iterator3.return) {
                    _iterator3.return();
                }
            } finally {
                if (_didIteratorError3) {
                    throw _iteratorError3;
                }
            }
        }
    }

    var ACTION_METHODS = (_ACTION_METHODS = {}, _defineProperty(_ACTION_METHODS, _constants.ADD_ACTION, addElement), _defineProperty(_ACTION_METHODS, _constants.UPDATE_ACTION, updateElement), _defineProperty(_ACTION_METHODS, _constants.REMOVE_ACTION, removeElement), _defineProperty(_ACTION_METHODS, _constants.REGISTER_STYLE, registerStyle), _ACTION_METHODS);

    var pause = function pause() {
        paused = true;
    };

    var resume = function resume() {
        paused = false;
        processPatch();
    };

    var queuePatch = function queuePatch(patchSet) {
        var finish = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

        if (!patchSet) {
            return;
        }
        pendingPatches.unshift(patchSet.map(annotatePatchAttributes));
        if (finish) {
            processPatch();
            pause();
        } else {
            (0, _trimkit.requestAnimationFrame)(processPatch, domRoot);
        }
    };

    var processPatch = function processPatch() {
        if (paused) {
            return;
        }

        var patchSet = pendingPatches.pop();
        if (patchSet) {
            // console.group('PatchSet');
            var patch = void 0;
            while (patch = patchSet.shift()) {
                // console.log(JSON.stringify(patch));
                ACTION_METHODS[patch[_constants.PATCH_TYPE]](patch);
            }

            while (postProcessing.length) {
                postProcessing.pop()();
            }
            // console.groupEnd('PatchSet');
        }
        if (pendingPatches.length) {
            (0, _trimkit.requestAnimationFrame)(processPatch, domRoot);
        } else {
            maintenance();
        }
    };

    (0, _trimkit.requestAnimationFrame)(processPatch, domRoot);

    return {
        queuePatch: queuePatch,
        pause: pause,
        resume: resume
    };
}