'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.getEventFlags = undefined;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

exports.nop = nop;
exports.traverseThunk = traverseThunk;
exports.shallowEqual = shallowEqual;
exports.memoize = memoize;
exports.isTag = isTag;

var _constants = require('./constants');

var _trimkit = require('trimkit');

function nop() {
    return;
};

function traverseThunk(thunk, errorDefault) {
    if (!(0, _trimkit.isFunction)(thunk)) {
        return thunk;
    }
    var result = thunk;
    do {
        try {
            result = result();
        } catch (e) {
            if ((0, _trimkit.isUndefined)(errorDefault)) {
                throw e;
            }
            console.error('Error rendering ' + result.name + ':', e);
            result = errorDefault;
        }
    } while ((0, _trimkit.isFunction)(result));
    return result;
}

// From https://github.com/gaearon/react-pure-render/ (MIT)
function shallowEqual(objA, objB) {
    if (objA === objB) {
        return true;
    }

    if ((typeof objA === 'undefined' ? 'undefined' : _typeof(objA)) !== 'object' || _trimkit.Null === objA || (typeof objB === 'undefined' ? 'undefined' : _typeof(objB)) !== 'object' || _trimkit.Null === objB) {
        return false;
    }

    var keysA = (0, _trimkit.ObjectKeys)(objA);
    var keysB = (0, _trimkit.ObjectKeys)(objB);

    if (keysA.length !== keysB.length) {
        return false;
    }

    // Test for A's keys different from B.
    var bHasOwnProperty = Object.prototype.hasOwnProperty.bind(objB);
    for (var i = 0; i < keysA.length; i++) {
        if (!bHasOwnProperty(keysA[i]) || objA[keysA[i]] !== objB[keysA[i]]) {
            return false;
        }
    }

    return true;
}

/*
* memoize.js
* by @philogb and @addyosmani
* with further optimizations by @mathias
* and @DmitryBaranovsk
* perf tests: http://bit.ly/q3zpG3
* Released under an MIT license.
*/

function memoize(fn) {
    fn.memoize = {};

    var wrappedFn = function wrappedFn() {
        var args = (0, _trimkit.asArray)(arguments),
            hash = '',
            i = args.length;
        var currentArg = _trimkit.Null;
        while (i--) {
            currentArg = args[i];
            hash += currentArg === Object(currentArg) ? JSON.stringify(currentArg) : currentArg;
        }
        return hash in fn.memoize ? fn.memoize[hash] : fn.memoize[hash] = fn.apply(this, args);
    };

    wrappedFn.clear = function () {
        fn.memoize = {};
    };

    return wrappedFn;
}

var getEventFlags = exports.getEventFlags = memoize(function getEventFlags(eventName, name, type) {
    var flags = [];

    if (name === 'input') {
        if (type === 'checkbox') {
            flags.push('target.checked');
        } else {
            flags.push('target.value');
        }
        if (['onkeyup', 'onkeydown', 'onkeypress'].indexOf(eventName) !== -1) {
            flags.push('keyCode', 'ctrlKey', 'altKey', 'shiftKey');
        }
    } else {
        flags.push('target.value');
    }
    return flags.join(',');
});

function isTag(obj) {
    return (0, _trimkit.isArray)(obj) && (0, _trimkit.isString)(obj[_constants.TAG_NAME]);
}